/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Mozilla MathML Editor.
 *
 * The Initial Developer of the Original Code is
 * MacKichan Software, Inc.
 * Portions created by the Initial Developer are Copyright (C) 2002
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Steve Swanson (steve.swanson@mackichan.com)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// MathML editing functions

var mmlns  = "http://www.w3.org/1998/Math/MathML";
var htmlns = "http://www.w3.org/1999/xhtml";

function newbox() {
  var box = doc.createElementNS(mmlns,"mi");
  box.appendChild(doc.createTextNode("\u25A1"));
  box.setAttribute("tempinput","true");
  return box;
}

function newrowbox() {
  var row = doc.createElementNS(mmlns,"mrow");
  row.appendChild(newbox());
  return row;
}

function zerowidthspace() {
  var mi = doc.createElementNS(mmlns,"mi");
  mi.appendChild(doc.createTextNode("\u200B"));
  return mi;
}

function newnone() {
  var none = doc.createElementNS(mmlns,"none");
  return none;
}

function findtagparent(node,tag) {
  if (!node || node.localName == tag)
    return node;
  else
    return findtagparent(node.parentNode,tag);
}

function findmathparent(node) {
  return findtagparent(node,"math");
}

function getselectednode(s) {
  var r = s.getRangeAt(0);
  var n = r.commonAncestorContainer; // get selection node
  if (n.nodeType == Node.TEXT_NODE)
    n = n.parentNode;
  return n;
}

function cant_nest_here(node) {
  if (findtagparent(node,"mi")
   || findtagparent(node,"mn")
   || findtagparent(node,"mo"))
    return true;
  else
    return false;
}

function nestinmath(node) {
  var d = doc.createDocumentFragment();
  var m = doc.createElementNS(mmlns,"math");
  m.setAttribute("xmlns",mmlns);
  var r = doc.createElementNS(mmlns,"mrow");
  r.appendChild(node); 
  m.appendChild(r); 
  d.appendChild(m);
  return d;
}

function nestinmrow(old) {
  if (old.localName == "mrow" || old.localName == "mo") {
    return old;
  } else {
    var row = doc.createElementNS(mmlns,"mrow");
    row.appendChild(old)
    return row;
  }
}

function insertbefore(loc,node) {
  var n = findmathparent(loc);
  var p = loc.parentNode;
  if (!n) {
    p.insertBefore(nestinmath(node),loc);
  } else {
    if (cant_nest_here(p)) {
      loc = p;
      p = loc.parentNode;
    }
    p.insertBefore(node,loc);
  }
}

function insertafter(loc,node) {
  var sib = loc.nextSibling;
  if (sib)
    loc.parentNode.insertBefore(node,sib);
  else
    loc.parentNode.appendChild(node);
}

function appendhere(loc,node) {
  var n = findmathparent(loc);
  if (!n) {
    var d = nestinmath(node);
    insertafter(loc,d);
    return;
  } else {
    var p = loc.parentNode;
    if (cant_nest_here(p))
      loc = p;
    insertafter(loc,node);
  }
}

function replacebox(loc,node) {
  var parent = loc.parentNode;
  var m = findmathparent(parent);
  if (parent.localName == "mi") {
    parent.parentNode.replaceChild(node,parent);
  } else {
    alert("impossible replacebox!");
  }
}

function inputboxselected(node) {
  if (node.nodeType == Node.TEXT_NODE
   && node.data == "\u25A1" && node.length == 1
   && node.parentNode.localName == "mi")
    return true;
  else if (node.localName == "mi")
    return inputboxselected(node.firstChild);
  else
    return false;
} 

function insertfragment(node) {
  var s = wind.getSelection();
  if (s) {
    var end = s.focusNode;
    var off = s.focusOffset;
    var sel_length = Math.abs(off-s.anchorOffset);
    if (s.anchorNode == end && inputboxselected(end)) {
      replacebox(end,node);
    } else if (end.nodeType == Node.TEXT_NODE && off > 0) {
      if (off == end.length ) {
        appendhere(end,node);
      } else {
        insertbefore(end.splitText(off),node);
      }
    } else if (end.localName == "mrow") { // for some reason, end.length is undefined at this point
      end.appendChild(node);
    } else {
      insertbefore(end,node);
    }
  }
}

function insertmath() {
  var d = doc.createDocumentFragment();
  var b = newbox();
  d.appendChild(b);
  insertfragment(d);
  caret_on_box(b);
}

function insertdisplay() {
  var math = doc.createElementNS(mmlns,"math");
  math.setAttribute("xmlns",mmlns);
  math.setAttribute("mode","display");
  var style = doc.createElementNS(mmlns,"mstyle");
  style.setAttribute("displaystyle","true");
  var rb = newrowbox();
  style.appendChild(rb);
  math.appendChild(style);
  var s = wind.getSelection();
  if (s) {
    var end = s.focusNode;
    var off = s.focusOffset;
    var p = end.parentNode;
    var n = findmathparent(end);
    if (n) {  // don't nest display inside math
      insertafter(n,math);
    } else if (end.nodeType == Node.TEXT_NODE && off > 0) {
      if (off == end.length ) {
        insertafter(end,math);
      } else {
        p.insertBefore(math,end.splitText(off));
      }
    } else {
      p.insertBefore(math,end);
    }
  }
  caret_on_box(rb);
}

function inserttag(tagname,op) {
  var d = doc.createDocumentFragment();
  var tag = doc.createElementNS(mmlns,tagname);
  var t = doc.createTextNode(op)
  tag.appendChild(t);
  d.appendChild(tag);
  insertfragment(d);
  caret_on_node(t);
}

function insertintegral() {
  inserttag("mo","\u222B");
}
function insertdoubleintegral() {
  inserttag("mo","\u222C");
}
function insertcontourintegral() {
  inserttag("mo","\u222E");
}
function insertsum() {
  inserttag("mo","\u2211");
}
function insertprod() {
  inserttag("mo","\u220F");
}
function insertdirectprod() {
  inserttag("mo","\u2297");
}
function inserttimes() {
  inserttag("mo","\u00D7");
}
function insertdivide() {
  inserttag("mo","\u00F7");
}
function insertplusminus() {
  inserttag("mo","\u00B1");
}
function insertcompose() {
  inserttag("mo","\u2218");
}
function insertcdot() {
  inserttag("mo","\u00B7");
}
function insertnabla() {
  inserttag("mo","\u2207");
}
function insertelement() {
  inserttag("mo","\u2208");
}
function insertsubset() {
  inserttag("mo","\u2282");
}
function insertalpha() {
  inserttag("mi","\u03B1");
}
function insertpi() {
  inserttag("mi","\u03C0");
}
function inserttheta() {
  inserttag("mi","\u03B8");
}
function insertinfinity() {
  inserttag("mi","\u221E");
}
function insertpartial() {
  inserttag("mi","\u2202");
}
function insertell() {
  inserttag("mi","\u2113");
}
function insertarrow() {
  inserttag("mo","\u2192");
}
function insertle() {
  inserttag("mo","\u2264");
}
function insertne() {
  inserttag("mo","\u2260");
}
function insertge() {
  inserttag("mo","\u2265");
}

function insertfence(left,right) {
  var d = doc.createDocumentFragment();
  var ml = doc.createElementNS(mmlns,"mo");
  var mr = doc.createElementNS(mmlns,"mo");
  ml.appendChild(doc.createTextNode(left));
  ml.setAttribute("stretchy","true");
  mr.appendChild(doc.createTextNode(right));
  mr.setAttribute("stretchy","true");
  var s = wind.getSelection();
  if (s.isCollapsed || !findmathparent(s.focusNode)) {
    d.appendChild(ml);
    var rb = newrowbox();
    d.appendChild(rb); 
    d.appendChild(mr);
    insertfragment(d);
    caret_on_box(rb);
  } else {
    var n = getselectednode(s);
    var old = n.cloneNode(true);
    d.appendChild(ml);
    d.appendChild(nestinmrow(old));
    d.appendChild(mr);
    n.parentNode.replaceChild(d,n);
  }
}

function insertbrackets() {
  insertfence("[","]");
}
function insertparens() {
  insertfence("(",")");
}
function insertbraces() {
  insertfence("{","}");
}

function caret_on_box(rb) {
  var c = rb.firstChild;
  while (c && c.nodeType != Node.TEXT_NODE)
    c = c.firstChild;
  var s = wind.getSelection();
  s.collapse(c,1);
}

// Actually, put caret at end of last leaf under node
function caret_on_node(node) {
  var c = node;
  while (c && c.nodeType != Node.TEXT_NODE)
    c = c.lastChild;
  var s = wind.getSelection();
  s.collapse(c,c.length);
}

function insertfractionex(attr,val) {
  var d = doc.createDocumentFragment();
  var mf = doc.createElementNS(mmlns,"mfrac");
  if (attr.length > 0)
    mf.setAttribute(attr,val);
  var s = wind.getSelection();
  if (s.isCollapsed || !findmathparent(s.focusNode)) {
    var rb = newrowbox();
    mf.appendChild(rb);
    mf.appendChild(newrowbox());
    d.appendChild(mf);
    insertfragment(d);
    caret_on_box(rb);
  } else {
    var n = getselectednode(s);
    var old = n.cloneNode(true);
    mf.appendChild(nestinmrow(old));
    mf.appendChild(newrowbox());
    d.appendChild(mf);
    n.parentNode.replaceChild(d,n);
  }
}

function insertfraction() {
  insertfractionex("","");
}

function insertfractionnoline() {
  insertfractionex("linethickness","0");
}

function insertfractionbevelled() {
  insertfractionex("bevelled","true");
}

function insertfractionthick() {
  insertfractionex("linethickness","thick");
}

function insertsqrt() {
  var d = doc.createDocumentFragment();
  var ms = doc.createElementNS(mmlns,"msqrt");
  var s = wind.getSelection();
  if (s.isCollapsed || !findmathparent(s.focusNode)) {
    var rb = newrowbox();
    ms.appendChild(rb);
    d.appendChild(ms);
    insertfragment(d);
    caret_on_box(rb);
  } else {
    var n = getselectednode(s);
    var old = n.cloneNode(true);
    ms.appendChild(nestinmrow(old));
    d.appendChild(ms);
    n.parentNode.replaceChild(d,n);
  }
}

function insertroot() {
  var d = doc.createDocumentFragment();
  var ms = doc.createElementNS(mmlns,"mroot");
  var s = wind.getSelection();
  if (s.isCollapsed || !findmathparent(s.focusNode)) {
    var rb = newrowbox();
    ms.appendChild(rb);
    ms.appendChild(newrowbox());
    d.appendChild(ms);
    insertfragment(d);
    caret_on_box(rb);
  } else {
    var n = getselectednode(s);
    var old = n.cloneNode(true);
    ms.appendChild(nestinmrow(old));
    ms.appendChild(newrowbox());
    d.appendChild(ms);
    n.parentNode.replaceChild(d,n);
  }
}

function insertname(n) {
  var d = doc.createDocumentFragment();
  var mi = doc.createElementNS(mmlns,"mi");
  var t = doc.createTextNode(n);
  mi.appendChild(t);
  mi.setAttribute("mathcolor","gray");
  d.appendChild(mi);
  insertfragment(d);
  var s = wind.getSelection();
  s.collapse(t,n.length);  // Put caret after insertion
}

function insertsin() {
  insertname("sin");
}
function insertcos() {
  insertname("cos");
}
function inserttan() {
  insertname("tan");
}
function insertln() {
  insertname("ln");
}
function insertexp() {
  insertname("exp");
}
function insertlim() {
  insertname("lim");
}

function isFirstChildOf(node,tag) {
  if (node.parentNode.localName != tag)
    return false;
  else if (node.previousSibling)
    return false;
  else
    return true;
}

// if we're the only child of an mrow, use that instead.
function useEnclosingmrow(node) {
  if (node.previousSibling || node.nextSibling ||
      node.parentNode.localName != "mrow")
    return node;
  else
    return node.parentNode;
}

function insertsup() {
  var d  = doc.createDocumentFragment();
  var ms = doc.createElementNS(mmlns,"msup");
  var s  = wind.getSelection();
  var n  = getselectednode(s);
  var ismo = (n.localName == "mo");
  var rb = newrowbox();  // new empty superscript
  var mr = useEnclosingmrow(n);
  if (isFirstChildOf(mr,"msub")) {  // msubsup looks better than nested msub/msup
    n = mr;
    var msub = n.parentNode;
    var base = n.cloneNode(true);
    var sub  = n.nextSibling.cloneNode(true);
    ms = doc.createElementNS(mmlns,"msubsup");
    ms.appendChild(base);
    ms.appendChild(sub);
    ms.appendChild(rb);
    d.appendChild(ms);
    msub.parentNode.replaceChild(d,msub);
  } else if (!ismo && (s.isCollapsed || !findmathparent(n))) {
    ms.appendChild(newrowbox());
    ms.appendChild(rb);
    d.appendChild(ms);
    insertfragment(d);
  } else {
    var old = n.cloneNode(true);
    ms.appendChild(nestinmrow(old));
    ms.appendChild(rb);
    d.appendChild(ms);
    n.parentNode.replaceChild(d,n);
  }
  caret_on_box(rb);
}

function insertsub() {
  var d  = doc.createDocumentFragment();
  var ms = doc.createElementNS(mmlns,"msub");
  var s  = wind.getSelection();
  var n  = getselectednode(s);
  var ismo = (n.localName == "mo");
  var rb = newrowbox();  // new empty subscript
  var mr = useEnclosingmrow(n);
  if (isFirstChildOf(mr,"msup")) {
    n = mr;
    var msup = n.parentNode;
    var base = n.cloneNode(true);
    var sup  = n.nextSibling.cloneNode(true);
    ms = doc.createElementNS(mmlns,"msubsup");
    ms.appendChild(base);
    ms.appendChild(rb);
    ms.appendChild(sup);
    d.appendChild(ms);
    msup.parentNode.replaceChild(d,msup);
  } else if (!ismo && (s.isCollapsed || !findmathparent(n))) {
    ms.appendChild(newrowbox());
    ms.appendChild(rb);
    d.appendChild(ms);
    insertfragment(d);
  } else {
    var old = n.cloneNode(true);
    ms.appendChild(nestinmrow(old));
    ms.appendChild(rb);
    d.appendChild(ms);
    n.parentNode.replaceChild(d,n);
  }
  caret_on_box(rb);
}

function insertsubsup() {
  var d = doc.createDocumentFragment();
  var ms = doc.createElementNS(mmlns,"msubsup");
  var s = wind.getSelection();
  var n  = getselectednode(s);
  var rb = newrowbox();  // new empty superscript
  var ismo = (n.localName == "mo");
  if (!ismo && (s.isCollapsed || !findmathparent(n))) {
    ms.appendChild(newrowbox());
    ms.appendChild(newrowbox());
    ms.appendChild(rb);
    d.appendChild(ms);
    insertfragment(d);
  } else {
    var old = n.cloneNode(true);
    ms.appendChild(nestinmrow(old));
    ms.appendChild(newrowbox());
    ms.appendChild(rb);
    d.appendChild(ms);
    n.parentNode.replaceChild(d,n);
  }
  caret_on_box(rb);
}

function inserttensor() {
  var d = doc.createDocumentFragment();
  var ms = doc.createElementNS(mmlns,"mmultiscripts");
  var mp = doc.createElementNS(mmlns,"mprescripts");
  var s = wind.getSelection();
  var n  = getselectednode(s);
  var rb = newrowbox();  // new empty subscript
  if (s.isCollapsed || !findmathparent(n)) {
    ms.appendChild(newrowbox());
    ms.appendChild(newrowbox());
    ms.appendChild(rb);
    ms.appendChild(mp);
    ms.appendChild(newrowbox());
    ms.appendChild(newrowbox());
    d.appendChild(ms);
    insertfragment(d);
  } else {
    var old = n.cloneNode(true);
    ms.appendChild(nestinmrow(old));
    ms.appendChild(newrowbox());
    ms.appendChild(rb);
    ms.appendChild(mp);
    ms.appendChild(newrowbox());
    ms.appendChild(newrowbox());
    d.appendChild(ms);
    n.parentNode.replaceChild(d,n);
  }
  caret_on_box(rb);
}

function insertmover() {
  var d = doc.createDocumentFragment();
  var mover = doc.createElementNS(mmlns,"mover");
  var s = wind.getSelection();
  var n  = getselectednode(s);
  var rb = newrowbox();  // new empty overscript
  var mr = useEnclosingmrow(n);
  if (isFirstChildOf(mr,"munder")) {
    n = mr;
    var munder = n.parentNode;
    var base   = n.cloneNode(true);
    var under  = n.nextSibling.cloneNode(true);
    ms = doc.createElementNS(mmlns,"munderover");
    ms.appendChild(base);
    ms.appendChild(under);
    ms.appendChild(rb);
    d.appendChild(ms);
    munder.parentNode.replaceChild(d,munder);
  } else if (s.isCollapsed || !findmathparent(n)) {
    mover.appendChild(newrowbox());
    mover.appendChild(rb);
    d.appendChild(mover);
    insertfragment(d);
  } else {
    var old = n.cloneNode(true);
    mover.appendChild(nestinmrow(old));
    mover.appendChild(rb);
    d.appendChild(mover);
    n.parentNode.replaceChild(d,n);
  }
  caret_on_box(rb);
}

function insertmunder() {
  var d = doc.createDocumentFragment();
  var munder = doc.createElementNS(mmlns,"munder");
  var s = wind.getSelection();
  var n  = getselectednode(s);
  var rb = newrowbox();  // new empty underscript
  var mr = useEnclosingmrow(n);
  if (isFirstChildOf(mr,"mover")) {
    n = mr;
    var mover = n.parentNode;
    var base  = n.cloneNode(true);
    var over  = n.nextSibling.cloneNode(true);
    ms = doc.createElementNS(mmlns,"munderover");
    ms.appendChild(base);
    ms.appendChild(rb);
    ms.appendChild(over);
    d.appendChild(ms);
    mover.parentNode.replaceChild(d,mover);
  } else if (s.isCollapsed || !findmathparent(n)) {
    munder.appendChild(newrowbox());
    munder.appendChild(rb);
    d.appendChild(munder);
    insertfragment(d);
  } else {
    n = getselectednode(s);
    var old = n.cloneNode(true);
    munder.appendChild(nestinmrow(old));
    munder.appendChild(rb);
    d.appendChild(munder);
    n.parentNode.replaceChild(d,n);
  }
  caret_on_box(rb);
}

function insertmunderover() {
  var d = doc.createDocumentFragment();
  var muo = doc.createElementNS(mmlns,"munderover");
  var rb = newrowbox();  // new empty overscript
  var s = wind.getSelection();
  if (s.isCollapsed || !findmathparent(s.focusNode)) {
    muo.appendChild(newrowbox());
    muo.appendChild(newrowbox());
    muo.appendChild(rb);
    d.appendChild(muo);
    insertfragment(d);
  } else {
    var n = getselectednode(s);
    var old = n.cloneNode(true);
    muo.appendChild(nestinmrow(old));
    muo.appendChild(newrowbox());
    muo.appendChild(rb);
    d.appendChild(muo);
    n.parentNode.replaceChild(d,n);
  }
  caret_on_box(rb);
}

function insertmatrix(rows,columns) {
  var d = doc.createDocumentFragment();
  var mt = doc.createElementNS(mmlns,"mtable");
  var mr = doc.createElementNS(mmlns,"mtr");
  var md = doc.createElementNS(mmlns,"mtd");
  md.appendChild(newrowbox());
  var i;
  for (i = 0; i < columns; i++)
    mr.appendChild(doc.importNode(md,true));
  for (i = 0; i < rows; i++)
    mt.appendChild(doc.importNode(mr,true));
  d.appendChild(mt);
  insertfragment(d);
  caret_on_box(mt);  // will descend to first box
}

// returns 0 for no siblings
function hassiblings(node) {
  var count = 0;
  if (node.previousSibling)
    count++;
  if (node.nextSibling)
    count++;
  return count;
}

function getsibling(node) {
  if (!node)
    return null;
  var sib = node.nextSibling;
  if (!sib)
    sib = node.previousSibling;
  return sib;
}

function deleteleft() {
  var s = wind.getSelection();
  if (!s.isCollapsed)
    deleteselection();
  else if (!findmathparent(s.focusNode))
    deletetext(s,false);
}

function deletetext(s,right) {
  var n = s.focusNode;
  if (n != s.anchorNode || n.nodeType != Node.TEXT_NODE)  // can't deal with inter-object selections yet
    return;
  var off = s.anchorOffset;
  var count = s.focusOffset - s.anchorOffset;
  if (s.isCollapsed)
  {
    count = 1;
    if (!right) {  // backspace
      if (off == 0)
        return;
      else
        --off;
    }
  } else if (count < 0) {
    count = -count;
    off = s.focusOffset;
  }
  var old = n.data;
  if (off + count <= n.length) {
    n.data = old.substr(0,off) + old.substr(off+count);
    s.collapse(n,off);
  } else {
    n.data = old.substr(0,off);
    s.collapse(n,off);
  }
}

function deleteselection() {
  var s = wind.getSelection();
  // A bit crude - mostly a problem with Mozilla selection behavior.
  var n  = getselectednode(s);
  if (findmathparent(n))
    deletemathnode(n);
  else
    deletetext(s,true);
}

function deletemathnode(n) {
  var mr = useEnclosingmrow(n);
  var p = mr.parentNode;
  var pname = p.localName;
  if (pname == "mrow") {       
    if (hassiblings(mr) == 0)   // this is the only child
      deletemathnode(p);
    else
      p.removeChild(mr);
  } else if (pname == "math") {   // nothing left, remove math
    p.parentNode.removeChild(p);
  } else if (pname == "msqrt") {
    if (inputboxselected(n))      // delete empty sqrt
      deletemathnode(p.parentNode);
    else                          // delete contents, put input box
      p.replaceChild(newrowbox(),mr)
  } else if (pname == "mfrac") {
    p.replaceChild(newrowbox(),mr)
  } else if (pname == "mroot") {
    p.replaceChild(newrowbox(),mr)       // if power, should revert to msqrt
  } else if (pname == "munder" || pname == "mover") {
    p.replaceChild(newrowbox(),mr)      // if box in label, should revert to base
  } else if (pname == "munderover") {
    p.replaceChild(newrowbox(),mr)      // if box in label, should revert to munder/mover
  } else if (pname == "msub"   || pname == "msup") {
    p.replaceChild(newrowbox(),mr)      // if box in label, should revert to base
  } else if (pname == "msubsup") {
    p.replaceChild(newrowbox(),mr)      // if box in script, should revert to msub/msup
  } else if (pname == "mmultiscripts") {
    if (inputboxselected(n) && mr.previousSibling) {  // box in script, delete position
      p.replaceChild(newnone(),mr);
    } else {
      p.replaceChild(newrowbox(),mr)
    }
  } else if (pname == "mtd") {
    p.replaceChild(newrowbox(),mr)
  } else if (pname == "mtr") {
    // must be mtd in mtr -- nothing to do (unless 1 column?)
  } else if (pname == "mtable") {
    // must be mtr in mtable -- need to delete row
    alert("Not implemented.");
  } else {
    alert("How to delete? " + mr.localName + " in " + pname);
  }
}

function sizesmall() {
  var s = wind.getSelection();
  var n  = getselectednode(s);
  if (findmathparent(n))
    n.setAttribute("mathsize","small");
}

function sizebig() {
  var s = wind.getSelection();
  var n  = getselectednode(s);
  if (findmathparent(n))
    n.setAttribute("mathsize","big");
}

function sizenormal() {
  var s = wind.getSelection();
  var n  = getselectednode(s);
  if (findmathparent(n))
    n.removeAttribute("mathsize");
}

// We use MathML 1.01 style attributes, since Mozilla doesn't support mathvariant.
// For now, at least, this is simpler anyway.

function stylebold(b) {
  var s = wind.getSelection();
  var n  = getselectednode(s);
  if (findmathparent(n)) {
    if (b)
      n.setAttribute("fontweight","bold");
    else
      n.removeAttribute("fontweight");
  }
}

function styleitalic(i) {
  var s = wind.getSelection();
  var n  = getselectednode(s);
  if (findmathparent(n)) {
    if (i)
      n.setAttribute("fontstyle","italic");
    else
      n.removeAttribute("fontstyle");
  }
}

function togglebold() {
  var s = wind.getSelection();
  var n  = getselectednode(s);
  if (findmathparent(n)) {
    if (n.getAttribute("fontweight") != "bold")
      n.setAttribute("fontweight","bold");
    else
      n.removeAttribute("fontweight");
  }
}

function toggleitalic() {
  var s = wind.getSelection();
  var n  = getselectednode(s);
  if (findmathparent(n)) {
    if (n.getAttribute("fontstyle") != "italic")
      n.setAttribute("fontstyle","italic");
    else
      n.removeAttribute("fontstyle");
  }
}

function setcolor(c) {
  var s = wind.getSelection();
  var n  = getselectednode(s);
  if (findmathparent(n)) {
    if (c)
      n.setAttribute("mathcolor",c);
    else
      n.removeAttribute("mathcolor");
  }
}

function setbackgroundcolor(c) {
  var s = wind.getSelection();
  var n  = getselectednode(s);
  if (findmathparent(n)) {
    if (c)
      n.setAttribute("mathbackground",c);
    else
      n.removeAttribute("mathbackground");
  }
}

// A raw attribute list.  Edit Attributes is more refined.
function viewattributes(n) {
  var node = n;
  if (node.nodeType == Node.TEXT_NODE)
    node = node.parentNode;
  if (findmathparent(node)) {
    var msg = "Node name: " + node.nodeName;
    msg += "\n\n";

    for (var i=0; i<node.attributes.length; i++) {
      var attr = node.attributes.item(i);
      msg += attr.nodeName;
      msg += '=';
      msg += attr.nodeValue;
      msg += '\n';
    }
    alert(msg);
  }
}

function movespace(n) {
  var event = doc.createEvent("KeyEvents");
  event.initKeyEvent("keypress",true,true,null,false,false,false,false,event.DOM_VK_RIGHT,0);
  n.dispatchEvent(event);
}

function inserttextchar(s,str) {
  var n = s.focusNode;
  var off = s.focusOffset;
  if (n.nodeType == Node.TEXT_NODE) {
    var old = n.data;
    n.data = old.substr(0,off) + str + old.substr(off);
    s.collapse(n,off+1);
  }
}

function insertspace(s) {
  inserttextchar(s," ");
}

function tabobject() {
  var s = wind.getSelection();
  var n = s.focusNode;
  if (!findmathparent(n))
    return;
  if (!s.isCollapsed) {
    s.collapseToStart();
    return;
  }
  
  // find enclosing object which responds to TAB
  for (var p = n; p.localName != "math"; p = p.parentNode) {
    if (p.localName == "mtable") {
      tabtable(s,n);
      return;
    } else if (p.localName == "mroot") {
      tabroot(s,n);
      return;
    } else if (p.localName == "mfrac") {
      tabfrac(s,n);
      return;    
    } else if (p.localName == "msubsup") {
      tabsubsup(s,n);
      return;    
    } else if (p.localName == "munderover") {
      tabunderover(s,n);
      return;
    }    
  }
}

function tabtable(s,n) {
  var mtd = findtagparent(n,"mtd");
  var sib = getnextmtd(mtd);
  if (sib) {
    caret_on_node(sib);
  } else {
    alert("lost in table");
  }
}

function getnextmtd(n) {
  if (!n || n.localName != "mtd")
    return null;
  var mtr = n.parentNode;
  var sib = n.nextSibling;
  if (!sib) {
    var nextrow = mtr.nextSibling;
    if (nextrow) {
      sib = nextrow.firstChild;
    }
    if (!sib) {
      var mtable = mtr.parentNode;
      mtr = mtable.firstChild;
      sib = mtr.firstChild;
    }
  }
  return sib; 
}

function tabfrac(s,n) {
  var me = n;
  while (me && me.parentNode.localName != "mfrac")
    me = me.parentNode;
  var sib = getsibling(me);
  if (sib) {
    caret_on_node(sib);
  } else {
    alert("lost in mfrac");
  }
}

function tabroot(s,n) {
  var me = n;
  while (me && me.parentNode.localName != "mroot")
    me = me.parentNode;
  var sib = getsibling(me);
  if (sib) {
    caret_on_node(sib);
  } else {
    alert("lost in mroot");
  }
}

function tabsubsup(s,n) {
  var me = n;
  while (me && me.parentNode.localName != "msubsup")
    me = me.parentNode;
  var sib = getsibling(me);
  if (sib) {
    caret_on_node(sib);
  } else {
    alert("lost in msubsup");
  }
}

function tabunderover(s,n) {
  var me = n;
  while (me && me.parentNode.localName != "munderover")
    me = me.parentNode;
  var sib = getsibling(me);
  if (sib) {
    caret_on_node(sib);
  } else {
    alert("lost in munderover");
  }
}

// since NodeIterator isn't implemented in Mozilla
function findnextsibling(node,parent) {
  for (var p = node; p && p != parent; p = p.parentNode)
    if (p.nextSibling)
      return p.nextSibling;
  return null;
}

function dosplit() {
  var s = wind.getSelection();
  var p = findtagparent(s.focusNode,"p");
  if (!p)
    return;  // I'm lost (until we have some more editing behavior specified)
  var oldempty = false;
  var movenode = null;
  var end = s.focusNode;
  var off = s.focusOffset;
  var mathnode = findmathparent(end);
  if (mathnode) {
    movenode = findnextsibling(mathnode,p);
  } else if (end.nodeType == Node.TEXT_NODE) {
    if (off == 0) {
      movenode = end;
    } else if (off == end.length) {
      movenode = findnextsibling(end,p);
    } else {
      movenode = end.splitText(off);
    }
  } else {
    movenode = end;
  }
  // split the tree above movenode (under p)
  var t, next;
  if (movenode) {
    for (var parent = movenode.parentNode;
         parent && parent != p;
         parent = movenode.parentNode) {
      var newnode = parent.cloneNode(false);  // shallow copy
      for (t = movenode; t; ) {
        next = t.nextSibling;
        newnode.appendChild(t.parentNode.removeChild(t));
        t = next;
      }
      insertafter(parent,newnode);
      movenode = parent.nextSibling;  // i.e. the new one
    }
  }
  var para = doc.createElementNS(htmlns,"p");
  if (movenode) {
    if (movenode == p.firstChild)
      oldempty = true;
    for (t = movenode; t; ) {
      next = t.nextSibling;
      para.appendChild(t.parentNode.removeChild(t));
      t = next;
    }
  } else {
    para.appendChild(doc.createTextNode("new paragraph"));  // localize!
  }
  insertafter(p,para);
  if (oldempty)
    p.appendChild(doc.createTextNode("blank paragraph"));  // localize!
}

function dospace() {
  var s = wind.getSelection();
  if (!s.isCollapsed) {
    s.collapseToStart();
  } else {
    if (findmathparent(s.focusNode))
      movespace(s.focusNode);
    else
      insertspace(s);
  }
}

function insertchar(code)
{
  var s = wind.getSelection();
  if (findmathparent(s.focusNode)) {
    switch (code) {
    case 33:   // '!'
    case 37:   // '%'
    case 38:   // '&'
    case 39:   // '''
    case 40:   // '('
    case 41:   // ')'
    case 42:   // '*'
    case 43:   // '+'
    case 44:   // ','
    case 45:   // '-'
    case 46:   // '.'
    case 47:   // '/'
    case 58:   // ':'
    case 59:   // ';'
    case 60:   // '<'
    case 61:   // '='
    case 62:   // '>'
    case 91:   // '['
    case 92:   // '\'
    case 93:   // ']'
    case 94:   // '^'
    case 123:  // '{'
    case 124:  // '|'
    case 125:  // '}'
    case 126:  // '~'
      inserttag("mo",String.fromCharCode(code));
      break;
    case 48:
    case 49:
    case 50:
    case 51:
    case 52:
    case 53:
    case 54:
    case 55:
    case 56:
    case 57:
      inserttag("mn",String.fromCharCode(code));
      break;
    default:
      inserttag("mi",String.fromCharCode(code));
      break;
    }
  } else {
    inserttextchar(s,String.fromCharCode(code));
  }
}
